from trl import SFTTrainer
from transformers import TrainingArguments, DataCollatorForSeq2Seq
from unsloth import is_bfloat16_supported
from transformers import EarlyStoppingCallback
from naima_lab.models.config import Config
from unsloth import FastLanguageModel
from unsloth.chat_templates import get_chat_template
from unsloth.chat_templates import standardize_sharegpt
from functools import partial
from unsloth.chat_templates import train_on_responses_only
from naima_lab.models.enums import SaveMethod
from datasets import Dataset
from typing import Any

class ClassificatorTrainLLM:
    def __init__(self, config: Config):
        self.config = config

    @staticmethod
    def formatting_prompts_func_instruct(tokenizer, examples):
        convos = examples["conversations"]
        texts = [
            tokenizer.apply_chat_template(
                convo, tokenize=False, add_generation_prompt=False
            )
            for convo in convos
        ]
        return {
            "text": texts,
        }

    def _get_dataset(self, dataset: list[dict], tokenizer: Any) -> Dataset:
        dataset = Dataset.from_dict({"conversations": dataset})
        
        tokenizer = get_chat_template(
            tokenizer,
            chat_template=self.config.chat_template,
        )
        dataset: Dataset = standardize_sharegpt(dataset)
        dataset = dataset.map(
            partial(self.formatting_prompts_func_instruct, tokenizer),
            batched=True,
        )

        dataset = dataset.shuffle(seed=42)
        return dataset


    def fine_tune_instruct(self, dataset: list[dict]) -> None:
                        
        model, tokenizer = FastLanguageModel.from_pretrained(
            model_name=self.config.model_name,
            max_seq_length=self.config.max_seq_length,
            dtype=(None),
            load_in_4bit=self.config.load_in_4bit,
        )
        model = FastLanguageModel.get_peft_model(
            model,
            r=self.config.unsloth.fast_language_model_config.r,  # Choose any number > 0 ! Suggested 8, 16, 32, 64, 128
            target_modules=self.config.unsloth.fast_language_model_config.target_modules,
            lora_alpha=self.config.unsloth.fast_language_model_config.lora_alpha,
            lora_dropout=self.config.unsloth.fast_language_model_config.lora_dropout,  # Supports any, but = 0 is optimized
            bias=self.config.unsloth.fast_language_model_config.bias,  # Supports any, but = "none" is optimized
            use_gradient_checkpointing=self.config.unsloth.fast_language_model_config.use_gradient_checkpointing,  # True or "unsloth" for very long context
            random_state=self.config.unsloth.fast_language_model_config.random_state,
            use_rslora=self.config.unsloth.fast_language_model_config.use_rslora,  # We support rank stabilized LoRA
            loftq_config=self.config.unsloth.fast_language_model_config.loftq_config,  # And LoftQ
        )
        tokenizer = get_chat_template(
            tokenizer,
            chat_template=self.config.chat_template,
        )
        dataset = self._get_dataset(dataset, tokenizer).select(range(10))
        if self.config.test_size == 0:
            train_dataset = dataset
            eval_dataset = None
        else:
            train_val_split = dataset.train_test_split(test_size=self.config.test_size)
            train_dataset = train_val_split["train"]
            eval_dataset = train_val_split["test"]
        if self.config.trl.callbacks.use_callbaks:
            callbacks = [
                    EarlyStoppingCallback(
                        early_stopping_patience=self.config.trl.callbacks.early_stopping_patience,
                        early_stopping_threshold=self.config.trl.callbacks.early_stopping_threshold,
                    )
                ]
        else:
            callbacks = []
        trainer = SFTTrainer(
            model=model,
            tokenizer=tokenizer,
            train_dataset=train_dataset,  # I changed this
            eval_dataset=eval_dataset,  # I added this
            dataset_text_field=self.config.trl.dataset_text_field,
            max_seq_length=self.config.max_seq_length,
            data_collator=DataCollatorForSeq2Seq(tokenizer=tokenizer),
            dataset_num_proc=self.config.trl.dataset_num_proc,
            packing=self.config.trl.packing,  # Can make training 5x faster for short sequences.
            callbacks=callbacks,
            args=TrainingArguments(
                run_name=self.config.trl.args.run_name,
                per_device_train_batch_size=self.config.trl.args.per_device_train_batch_size,
                gradient_accumulation_steps=self.config.trl.args.gradient_accumulation_steps,
                warmup_steps=self.config.trl.args.warmup_steps,
                num_train_epochs=self.config.trl.args.num_train_epochs,
                do_eval=self.config.trl.args.do_eval,
                eval_steps=self.config.trl.args.eval_steps,
                max_steps=self.config.trl.args.max_steps,
                learning_rate=self.config.trl.args.learning_rate,
                fp16=not is_bfloat16_supported(),
                bf16=is_bfloat16_supported(),
                logging_steps=self.config.trl.args.logging_steps,
                optim=self.config.trl.args.optim,
                weight_decay=self.config.trl.args.weight_decay,
                lr_scheduler_type=self.config.trl.args.lr_scheduler_type,
                seed=self.config.trl.args.seed,
                logging_first_step=self.config.trl.args.logging_first_step,
                eval_strategy=self.config.trl.args.eval_strategy,
                save_strategy=self.config.trl.args.save_strategy,
                output_dir=self.config.trl.args.output_dir,
                report_to=self.config.trl.args.report_to,  # Use this for WandB etc
                load_best_model_at_end=self.config.trl.args.load_best_model_at_end,
            ),
        )
        trainer = train_on_responses_only(
            trainer,
            instruction_part=self.config.train_on_responses_only.instruction_part,
            response_part=self.config.train_on_responses_only.response_part,
        )

        trainer.train()
        if self.config.save_method == SaveMethod.ADAPTERS:
            model.save_pretrained(self.config.model_save_model)  # Local saving
            tokenizer.save_pretrained(self.config.model_save_model)  # Local saving            
        elif self.config.save_method == SaveMethod.FULL:
            model.save_pretrained(self.config.model_save_model,tokenizer, save_method = "merged_16bit",)
        else:
            raise ValueError(f"Save method {self.config.save_method} not supported.")


from typing import Optional

from pydantic import (
    BaseModel,
    Field,
)
import yaml
import os
from naima_lab.models.enums import SaveMethod 

class CallbackConfig(BaseModel):
    use_callbaks: bool = Field(..., description="Whether to use callbacks.")
    early_stopping_patience: int = Field(
        ..., description="Number of patience steps for early stopping."
    )
    early_stopping_threshold: float = Field(
        ..., description="Threshold for early stopping."
    )


class TrainerArgs(BaseModel):
    run_name: str = Field(..., description="Name of the training run.")
    per_device_train_batch_size: int = Field(
        ..., description="Batch size per device during training."
    )
    gradient_accumulation_steps: int = Field(
        ..., description="Number of gradient accumulation steps."
    )
    warmup_steps: int = Field(..., description="Number of warmup steps.")
    num_train_epochs: int = Field(..., description="Total number of training epochs.")
    max_steps: int = Field(..., description="Maximum number of steps.")
    do_eval: bool = Field(..., description="Whether to perform evaluation.")
    eval_steps: int = Field(..., description="Number of steps between evaluations.")
    learning_rate: float = Field(..., description="Learning rate.")
    logging_steps: int = Field(..., description="Number of steps between logging.")
    optim: str = Field(..., description="Optimizer to use.")
    weight_decay: float = Field(..., description="Weight decay for the optimizer.")
    lr_scheduler_type: str = Field(..., description="Type of learning rate scheduler.")
    seed: int = Field(..., description="Random seed.")
    logging_first_step: bool = Field(..., description="Log the first step.")
    eval_strategy: str = Field(..., description="Evaluation strategy.")
    save_strategy: str = Field(..., description="Model save strategy.")
    output_dir: str = Field(..., description="Output directory for the model.")
    report_to: str = Field(..., description="Reporting target.")
    load_best_model_at_end: bool = Field(
        ..., description="Whether to load the best model at the end."
    )


class TrlConfig(BaseModel):
    dataset_text_field: str = Field(
        ..., description="Field in the dataset that contains text."
    )
    dataset_num_proc: int = Field(
        ..., description="Number of processes for dataset loading."
    )
    packing: bool = Field(..., description="Whether to pack the dataset.")
    callbacks: CallbackConfig
    args: TrainerArgs


class FastLanguageModelConfig(BaseModel):
    r: int
    target_modules: list[str]
    lora_alpha: int
    lora_dropout: float
    bias: str
    use_gradient_checkpointing: bool | str
    random_state: int
    use_rslora: bool
    loftq_config: Optional[dict]

class UnslothConfig(BaseModel):
    fast_language_model_config: FastLanguageModelConfig

class InstructionConfig(BaseModel):
    instruction_part: str = Field(..., description="Instruction part.")
    response_part: str = Field(..., description="Response part.")

class Config(BaseModel):
    model_name: str = Field(..., description="Name of the model.")
    chat_template: str = Field(..., description="ChatTemplate name.")
    max_seq_length: int = Field(..., description="Maximum sequence length.")
    load_in_4bit: bool = Field(
        ..., description="Whether to load the model in 4-bit precision."
    )
    test_size: float = Field(..., description="Test size.")
    save_method: SaveMethod = Field(..., description="Method to save the model. ")
    model_save_model: str = Field(..., description="Name of the model to save.")
    unsloth: UnslothConfig = Field(..., description="Unsloth configuration.")
    trl: TrlConfig = Field(..., description="TRL configuration.")
    train_on_responses_only: InstructionConfig = Field(..., description="Whether to train on responses only.")
    @staticmethod
    def get_config(
        config_path: str = os.path.join(
            os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            ),
            "config",
            "config.yml",
        ),
    ) -> "Config":
        with open(config_path, "r", encoding="utf-8") as file:
            config: dict = yaml.safe_load(file)
            config["file_path"] = config_path
        return Config(**config)


import logging

from langchain_openai import ChatOpenAI
from langchain_openai import OpenAIEmbeddings
from naima_lab.models.enums import LLMClients, EmbeddingsClients
from langchain_ollama.llms import OllamaLLM
from langchain_ollama import OllamaEmbeddings



class InterfaceLangChain:

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    @classmethod
    def get_chat_llm(self, client: LLMClients = LLMClients.ollama, 
                model: str = "llama3.2:1b",
                max_tokens: int = 4096, 
                temperature: float = 0,
                seed: int = 42,
                **args) -> ChatOpenAI | OllamaLLM:
        if client == LLMClients.ollama:
            llm = OllamaLLM(
                model=model,
                num_predict=max_tokens,
                temperature=temperature,
                seed=seed,
                **args
            )
        elif client == LLMClients.openai:
            llm = ChatOpenAI(
                model=model,
                max_tokens=max_tokens,
                temperature=temperature,
                seed=seed,
                **args
            )
        else:
            raise Exception("The client is not permitted.")
        
        return llm

    @classmethod
    def get_embeddings(self, client: EmbeddingsClients = EmbeddingsClients.ollama, 
                       model: str = "nomic-embed-text:latest",
                    **args) -> OpenAIEmbeddings | OllamaEmbeddings:
        if client == EmbeddingsClients.openai:
            embeddings = OpenAIEmbeddings(
                model=model,
                **args
            )
        elif client == EmbeddingsClients.ollama:
            embeddings = OllamaEmbeddings(
                model=model,
                **args
            )
        else:
            raise Exception("The client is not permitted.")            
        return embeddings


import typer
import logging
from typing_extensions import Annotated
import os
import json

from naima_lab.models.config import Config

logger = logging.getLogger(__name__)

app = typer.Typer()


@app.command()
def llm_sft_instruct(
    config_file: Annotated[
        str, typer.Option(help="This is the path to the config file")
    ] = os.path.join("config", "config_llama3.2_1B_fast.yml"),
    dataset_path: Annotated[
        str, typer.Option(help="This is the path to the dataset.")
    ] = os.path.join("data", "raw", "FineTome-100k.json"),
):
    """Fine-tune the LLM with the given dataset.
The dataset should be in the following format:
1[
[{
    "from": "human",
    "value": "Can you help me make pasta carbonara?"
},
{
    "from": "assistant",
    "value": "Would you like the traditional Roman recipe, or a simpler version?"
},..],
[{
    "from": "human",
    "value": "The traditional version please"
},
{
    "from": "assistant",
    "value": "The authentic Roman carbonara uses just a few ingredients: pasta, guanciale, eggs, Pecorino Romano, and black pepper. Would you like the detailed recipe?"
}]
]
}
    
    """
    from naima_lab.nlp.llm.train import ClassificatorTrainLLM
    
    try:
        dataset = json.load(open(dataset_path, "r"))
    except FileNotFoundError:
        raise ValueError("Dataset not found")
    config = Config.get_config(config_file)
    trainer = ClassificatorTrainLLM(config=config)
    trainer.fine_tune_instruct(dataset=dataset)



from enum import Enum


class LLMClients(str, Enum):
    ollama = "ollama"
    openai = "openai"


class EmbeddingsClients(str, Enum):
    openai = "openai"
    ollama = "ollama"

class ModelType(str, Enum):
    base = "base"
    instruct = "instruct"
    
class AzureSearchMode(str, Enum):
    TEXT = 'text'
    VECTOR = 'vector'
    HYBRID = 'hybrid'
    SEMANTIC = 'semantic'
    HYBRID_SEMANTIC = 'hybrid_semantic'    
    
    
class SaveMethod(str, Enum):
    ADAPTERS = "adapters"
    FULL = "full"